# -*- coding: utf-8 -*-

import time
import logging
from functools import wraps
from contextlib import contextmanager
from .connection_interface import ConnectionInterface
from ..query.grammars.grammar import QueryGrammar
from ..query import QueryBuilder
from ..query.expression import QueryExpression
from ..query.processors.processor import QueryProcessor
from ..schema.builder import SchemaBuilder
from ..dbal.schema_manager import SchemaManager
from ..exceptions.query import QueryException


query_logger = logging.getLogger("orator.connection.queries")
connection_logger = logging.getLogger("orator.connection")


def run(wrapped):
    """
    Special decorator encapsulating query method.
    """

    @wraps(wrapped)
    def _run(self, query, bindings=None, *args, **kwargs):
        self._reconnect_if_missing_connection()

        start = time.time()
        try:
            result = wrapped(self, query, bindings, *args, **kwargs)
        except Exception as e:
            result = self._try_again_if_caused_by_lost_connection(
                e, query, bindings, wrapped
            )

        t = self._get_elapsed_time(start)
        self.log_query(query, bindings, t)

        return result

    return _run


class Connection(ConnectionInterface):

    name = None

    def __init__(
        self,
        connection,
        database="",
        table_prefix="",
        config=None,
        builder_class=QueryBuilder,
        builder_default_kwargs=None,
    ):
        """
        :param connection: A dbapi connection instance
        :type connection: Connector

        :param database: The database name
        :type database: str

        :param table_prefix: The table prefix
        :type table_prefix: str

        :param config: The connection configuration
        :type config: dict
        """
        self._connection = connection
        self._cursor = None

        self._read_connection = None

        self._database = database

        if table_prefix is None:
            table_prefix = ""

        self._table_prefix = table_prefix

        if config is None:
            config = {}

        self._config = config

        self._reconnector = None

        self._transactions = 0

        self._lost_connection = False

        self._pretending = False

        self._builder_class = builder_class

        if builder_default_kwargs is None:
            builder_default_kwargs = {}

        self._builder_default_kwargs = builder_default_kwargs

        self._logging_queries = config.get("log_queries", False)
        self._logged_queries = []

        # Setting the marker based on config
        self._marker = None
        if self._config.get("use_qmark"):
            self._marker = "?"

        self._query_grammar = self.get_default_query_grammar()

        self._schema_grammar = None

        self._post_processor = self.get_default_post_processor()

        self._server_version = None

        self.use_default_query_grammar()

    def use_default_query_grammar(self):
        self._query_grammar = self.get_default_query_grammar()

    def get_default_query_grammar(self):
        return QueryGrammar()

    def use_default_schema_grammar(self):
        self._schema_grammar = self.get_default_schema_grammar()

    def get_default_schema_grammar(self):
        pass

    def use_default_post_processor(self):
        self._post_processor = self.get_default_post_processor()

    def get_default_post_processor(self):
        return QueryProcessor()

    def get_database_platform(self):
        return self._connection.get_database_platform()

    def get_schema_builder(self):
        """
        Retturn the underlying schema builder.

        :rtype: orator.schema.builder.SchemaBuilder
        """
        if not self._schema_grammar:
            self.use_default_schema_grammar()

        return SchemaBuilder(self)

    def table(self, table):
        """
        Begin a fluent query against a database table

        :param table: The database table
        :type table: str

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        query = self.query()

        return query.from_(table)

    def query(self):
        """
        Begin a fluent query

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        query = self._builder_class(
            self,
            self._query_grammar,
            self._post_processor,
            **self._builder_default_kwargs
        )

        return query

    def raw(self, value):
        return QueryExpression(value)

    def select_one(self, query, bindings=None):
        if bindings is None:
            bindings = {}

        records = self.select(query, bindings)

        if len(records):
            return records[1]

        return None

    def select_from_write_connection(self, query, bindings=None):
        if bindings is None:
            bindings = {}

        return self.select(query, bindings, False)

    @run
    def select(self, query, bindings=None, use_read_connection=True):
        if self.pretending():
            return []

        bindings = self.prepare_bindings(bindings)
        cursor = self._get_cursor_for_select(use_read_connection)
        cursor.execute(query, bindings)

        return cursor.fetchall()

    def select_many(
        self, size, query, bindings=None, use_read_connection=True, abort=False
    ):
        if self.pretending():
            yield []
        else:
            bindings = self.prepare_bindings(bindings)
            cursor = self._get_cursor_for_select(use_read_connection)

            try:
                cursor.execute(query, bindings)
            except Exception as e:
                if self._caused_by_lost_connection(e) and not abort:
                    self.reconnect()

                    for results in self.select_many(
                        size, query, bindings, use_read_connection, True
                    ):
                        yield results
                else:
                    raise
            else:
                results = cursor.fetchmany(size)
                while results:
                    yield results

                    results = cursor.fetchmany(size)

    def _get_cursor_for_select(self, use_read_connection=True):
        if use_read_connection:
            self._cursor = self.get_read_connection().cursor()
        else:
            self._cursor = self.get_connection().cursor()

        return self._cursor

    def insert(self, query, bindings=None):
        return self.statement(query, bindings)

    def update(self, query, bindings=None):
        return self.affecting_statement(query, bindings)

    def delete(self, query, bindings=None):
        return self.affecting_statement(query, bindings)

    @run
    def statement(self, query, bindings=None):
        if self.pretending():
            return True

        bindings = self.prepare_bindings(bindings)

        return self._new_cursor().execute(query, bindings)

    @run
    def affecting_statement(self, query, bindings=None):
        if self.pretending():
            return True

        bindings = self.prepare_bindings(bindings)

        cursor = self._new_cursor()
        cursor.execute(query, bindings)

        return cursor.rowcount

    def _new_cursor(self):
        self._cursor = self.get_connection().cursor()

        return self._cursor

    def get_cursor(self):
        return self._cursor

    @run
    def unprepared(self, query):
        if self.pretending():
            return True

        return bool(self.get_connection().execute(query))

    def prepare_bindings(self, bindings):
        if bindings is None:
            return []

        return bindings

    @contextmanager
    def transaction(self):
        self.begin_transaction()

        try:
            yield self
        except Exception as e:
            self.rollback()
            raise

        try:
            self.commit()
        except Exception:
            self.rollback()
            raise

    def begin_transaction(self):
        self._transactions += 1

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
        else:
            self._transactions -= 1

    def transaction_level(self):
        return self._transactions

    @contextmanager
    def pretend(self):
        self._logged_queries = []

        self._pretending = True

        try:
            yield self
        except Exception:
            self._pretending = False

        self._pretending = False

    def _try_again_if_caused_by_lost_connection(
        self, e, query, bindings, callback, *args, **kwargs
    ):
        if self._caused_by_lost_connection(e):

            self.reconnect()

            return callback(self, query, bindings, *args, **kwargs)

        raise QueryException(query, bindings, e)

    def _caused_by_lost_connection(self, e):
        message = str(e).lower()

        for s in [
            "server has gone away",
            "no connection to the server",
            "lost connection",
            "is dead or not enabled",
            "error while sending",
            "decryption failed or bad record mac",
            "server closed the connection unexpectedly",
            "ssl connection has been closed unexpectedly",
            "error writing data to the connection",
            "connection timed out",
            "resource deadlock avoided",
            "connection already closed",
            "eof detected",
        ]:
            if s in message:
                self._lost_connection = True
                return True

        return False

    def disconnect(self):
        connection_logger.debug("%s is disconnecting" % self.__class__.__name__)
        if self._connection:
            self._connection.close()

        if self._read_connection and self._connection != self._read_connection:
            self._read_connection.close()

        self.set_connection(None).set_read_connection(None)

        connection_logger.debug("%s disconnected" % self.__class__.__name__)

    def reconnect(self):
        connection_logger.debug("%s is reconnecting" % self.__class__.__name__)
        if self._reconnector is not None and callable(self._reconnector):
            return self._reconnector(self)

        raise Exception("Lost connection and no reconnector available")

    def _reconnect_if_missing_connection(self):
        if self.get_connection() is None or self.get_read_connection() is None:
            self.reconnect()

    def log_query(self, query, bindings, time_=None):
        if self.pretending():
            self._logged_queries.append(self._get_cursor_query(query, bindings))

        if not self._logging_queries:
            return

        query = self._get_cursor_query(query, bindings)

        if query:
            log = "Executed %s" % (query,)

            if time_:
                log += " in %sms" % time_

            query_logger.debug(
                log, extra={"query": query, "bindings": bindings, "elapsed_time": time_}
            )

    def _get_elapsed_time(self, start):
        return round((time.time() - start) * 1000, 2)

    def _get_cursor_query(self, query, bindings):
        if self._pretending:
            return query, bindings

        return query, bindings

    def get_logged_queries(self):
        return self._logged_queries

    def get_connection(self):
        return self._connection

    def get_read_connection(self):
        if self._transactions >= 1:
            return self.get_connection()

        if self._read_connection is not None:
            return self._read_connection

        return self._connection

    def set_connection(self, connection):
        if self._transactions >= 1 and not self._lost_connection:
            raise RuntimeError(
                "Can't swap dbapi connection" "while within transaction."
            )

        if self._lost_connection:
           self._lost_connection = False

        self._connection = connection

        return self

    def set_read_connection(self, connection):
        self._read_connection = connection

        return self

    def set_reconnector(self, reconnector):
        self._reconnector = reconnector

        return self

    def get_name(self):
        return self._config.get("name")

    def get_config(self, option):
        return self._config.get(option)

    def get_query_grammar(self):
        return self._query_grammar

    def set_query_grammar(self, grammar):
        self._query_grammar = grammar

    def get_schema_grammar(self):
        return self._schema_grammar

    def set_schema_grammar(self, grammar):
        self._schema_grammar = grammar

    def get_post_processor(self):
        """
        Get the query post processor used by the connection

        :return: The query post processor
        :rtype: QueryProcessor
        """
        return self._post_processor

    def set_post_processor(self, processor):
        """
        Set the query post processor used by the connection

        :param processor: The query post processor
        :type processor: QueryProcessor
        """
        self._post_processor = processor

    def pretending(self):
        return self._pretending

    def enable_query_log(self):
        self._logging_queries = True

    def disable_query_log(self):
        self._logging_queries = False

    def logging(self):
        return self._logging_queries

    def get_database_name(self):
        return self._database

    def get_table_prefix(self):
        return self._table_prefix

    def set_table_prefix(self, prefix):
        self._table_prefix = prefix

        self.get_query_grammar().set_table_prefix(prefix)

    def with_table_prefix(self, grammar):
        grammar.set_table_prefix(self._table_prefix)

        return grammar

    def get_column(self, table, column):
        schema = self.get_schema_manager()

        return schema.list_table_details(table).get_column(column)

    def get_schema_manager(self):
        return SchemaManager(self)

    def get_params(self):
        return self._connection.get_params()

    def get_marker(self):
        return self._marker

    def set_builder_class(self, klass, default_kwargs=None):
        self._builder_class = klass

        if default_kwargs is not None:
            self._builder_default_kwargs = default_kwargs

        return self

    def __enter__(self):
        self._reconnect_if_missing_connection()
        self.begin_transaction()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            try:
                self.commit()
            except Exception:
                self.rollback()
                raise
        else:
            self.rollback()
            raise (exc_type, exc_val, exc_tb)

    @property
    def server_version(self):
        if self._server_version is None:
            self._server_version = self.get_server_version()

        return self._server_version

    def get_server_version(self):
        return self._connection.get_server_version()



# -*- coding: utf-8 -*-

from orator.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class StatusCommand(BaseCommand):
    """
    Show a list of migrations up/down.

    migrate:status
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
    """

    def handle(self):
        """
        Executes the command.
        """
        database = self.option("database")

        self.resolver.set_default_connection(database)

        repository = DatabaseMigrationRepository(self.resolver, "migrations")

        migrator = Migrator(repository, self.resolver)

        if not migrator.repository_exists():
            return self.error("No migrations found")

        self._prepare_database(migrator, database)

        path = self.option("path")

        if path is None:
            path = self._get_migration_path()

        ran = migrator.get_repository().get_ran()

        migrations = []
        for migration in migrator._get_migration_files(path):
            if migration in ran:
                migrations.append(["<fg=cyan>%s</>" % migration, "<info>Yes</>"])
            else:
                migrations.append(["<fg=cyan>%s</>" % migration, "<fg=red>No</>"])

        if migrations:
            table = self.table(["Migration", "Ran?"], migrations)
            table.render()
        else:
            return self.error("No migrations found")

        for note in migrator.get_notes():
            self.line(note)

    def _prepare_database(self, migrator, database):
        migrator.set_connection(database)



# -*- coding: utf-8 -*-

BLANK_STUB = """from orator.migrations import Migration


class DummyClass(Migration):

    def up(self):
        \"\"\"
        Run the migrations.
        \"\"\"
        pass

    def down(self):
        \"\"\"
        Revert the migrations.
        \"\"\"
        pass
"""

CREATE_STUB = """from orator.migrations import Migration


class DummyClass(Migration):

    def up(self):
        \"\"\"
        Run the migrations.
        \"\"\"
        with self.schema.create('dummy_table') as table:
            table.increments('id')
            table.timestamps()

    def down(self):
        \"\"\"
        Revert the migrations.
        \"\"\"
        self.schema.drop('dummy_table')
"""

UPDATE_STUB = """from orator.migrations import Migration


class DummyClass(Migration):

    def up(self):
        \"\"\"
        Run the migrations.
        \"\"\"
        with self.schema.table('dummy_table') as table:
            pass

    def down(self):
        \"\"\"
        Revert the migrations.
        \"\"\"
        with self.schema.table('dummy_table') as table:
            pass
"""



# -*- coding: utf-8 -*-

from pygments.formatter import Formatter
from pygments.token import (
    Keyword,
    Name,
    Comment,
    String,
    Error,
    Number,
    Operator,
    Generic,
    Token,
    Whitespace,
)
from pygments.util import get_choice_opt


COMMAND_COLORS = {
    Token: ("", ""),
    Whitespace: ("fg=white", "fg=black;options=bold"),
    Comment: ("fg=white", "fg=black;options=bold"),
    Comment.Preproc: ("fg=cyan", "fg=cyan;options=bold"),
    Keyword: ("fg=blue", "fg=blue;options=bold"),
    Keyword.Type: ("fg=cyan", "fg=cyan;options=bold"),
    Operator.Word: ("fg=magenta", "fg=magenta;options=bold"),
    Name.Builtin: ("fg=cyan", "fg=cyan;options=bold"),
    Name.Function: ("fg=green", "fg=green;option=bold"),
    Name.Namespace: ("fg=cyan;options=underline", "fg=cyan;options=bold,underline"),
    Name.Class: ("fg=green;options=underline", "fg=green;options=bold,underline"),
    Name.Exception: ("fg=cyan", "fg=cyan;options=bold"),
    Name.Decorator: ("fg=black;options=bold", "fg=white"),
    Name.Variable: ("fg=red", "fg=red;options=bold"),
    Name.Constant: ("fg=red", "fg=red;options=bold"),
    Name.Attribute: ("fg=cyan", "fg=cyan;options=bold"),
    Name.Tag: ("fg=blue;options=bold", "fg=blue;options=bold"),
    String: ("fg=yellow", "fg=yellow"),
    Number: ("fg=blue", "fg=blue;options=bold"),
    Generic.Deleted: ("fg=red;options=bold", "fg=red;options=bold"),
    Generic.Inserted: ("fg=green", "fg=green;options=bold"),
    Generic.Heading: ("options=bold", "option=bold"),
    Generic.Subheading: ("fg=magenta;options=bold", "fg=magenta;options=bold"),
    Generic.Prompt: ("options=bold", "options=bold"),
    Generic.Error: ("fg=red;options=bold", "fg=red;options=bold"),
    Error: ("fg=red;options=bold,underline", "fg=red;options=bold,underline"),
}


class CommandFormatter(Formatter):
    r"""
    Format tokens with Cleo color sequences, for output in a text console.
    Color sequences are terminated at newlines, so that paging the output
    works correctly.

    The `get_style_defs()` method doesn't do anything special since there is
    no support for common styles.

    Options accepted:

    `bg`
        Set to ``"light"`` or ``"dark"`` depending on the terminal's background
        (default: ``"light"``).

    `colorscheme`
        A dictionary mapping token types to (lightbg, darkbg) color names or
        ``None`` (default: ``None`` = use builtin colorscheme).

    `linenos`
        Set to ``True`` to have line numbers on the terminal output as well
        (default: ``False`` = no line numbers).
    """
    name = "Command"
    aliases = ["command"]
    filenames = []

    def __init__(self, **options):
        Formatter.__init__(self, **options)
        self.darkbg = (
            get_choice_opt(options, "bg", ["light", "dark"], "light") == "dark"
        )
        self.colorscheme = options.get("colorscheme", None) or COMMAND_COLORS
        self.linenos = options.get("linenos", False)
        self._lineno = 0

    def format(self, tokensource, outfile):
        return Formatter.format(self, tokensource, outfile)

    def _write_lineno(self, outfile):
        self._lineno += 1
        outfile.write("%s%04d: " % (self._lineno != 1 and "\n" or "", self._lineno))

    def _get_color(self, ttype):
        # self.colorscheme is a dict containing usually generic types, so we
        # have to walk the tree of dots.  The base Token type must be a key,
        # even if it's empty string, as in the default above.
        colors = self.colorscheme.get(ttype)
        while colors is None:
            ttype = ttype.parent
            colors = self.colorscheme.get(ttype)
        return colors[self.darkbg]

    def format_unencoded(self, tokensource, outfile):
        if self.linenos:
            self._write_lineno(outfile)

        for ttype, value in tokensource:
            color = self._get_color(ttype)

            for line in value.splitlines(True):
                if color:
                    outfile.write("<%s>%s</>" % (color, line.rstrip("\n")))
                else:
                    outfile.write(line.rstrip("\n"))
                if line.endswith("\n"):
                    if self.linenos:
                        self._write_lineno(outfile)
                    else:
                        outfile.write("\n")

        if self.linenos:
            outfile.write("\n")



# -*- coding: utf-8 -*-

import re
import binascii
from ..utils import encode


class AbstractAsset(object):

    _name = None

    _namespace = None

    _quoted = False

    def _set_name(self, name):
        """
        Sets the name of this asset.

        :param name: The name of the asset
        :type name: str
        """
        if self._is_identifier_quoted(name):
            self._quoted = True
            name = self._trim_quotes(name)

        if "." in name:
            parts = name.split(".", 1)
            self._namespace = parts[0]
            name = parts[1]

        self._name = name

    def _is_in_default_namespace(self, default_namespace):
        return self._namespace == default_namespace or self._namespace is None

    def get_namespace_name(self):
        return self._namespace

    def get_shortest_name(self, default_namespace):
        shortest_name = self.get_name()
        if self._namespace == default_namespace:
            shortest_name = self._name

        return shortest_name.lower()

    def get_full_qualified_name(self, default_namespace):
        name = self.get_name()
        if not self._namespace:
            name = default_namespace + "." + name

        return name.lower()

    def is_quoted(self):
        return self._quoted

    def _is_identifier_quoted(self, identifier):
        return len(identifier) > 0 and (
            identifier[0] == "`" or identifier[0] == '"' or identifier[0] == "["
        )

    def _trim_quotes(self, identifier):
        return re.sub('[`"\[\]]', "", identifier)

    def get_name(self):
        if self._namespace:
            return self._namespace + "." + self._name

        return self._name

    def get_quoted_name(self, platform):
        keywords = platform.get_reserved_keywords_list()
        parts = self.get_name().split(".")
        for k, v in enumerate(parts):
            if self._quoted or keywords.is_keyword(v):
                parts[k] = platform.quote_identifier(v)

        return ".".join(parts)

    def _generate_identifier_name(self, columns, prefix="", max_size=30):
        """
        Generates an identifier from a list of column names obeying a certain string length.
        """
        hash = ""
        for column in columns:
            hash += "%x" % binascii.crc32(encode(str(column)))

        return (prefix + "_" + hash)[:max_size]



# -*- coding: utf-8 -*-

from collections import OrderedDict
from .table import Table
from .identifier import Identifier


class TableDiff(object):
    def __init__(
        self,
        table_name,
        added_columns=None,
        changed_columns=None,
        removed_columns=None,
        added_indexes=None,
        changed_indexes=None,
        removed_indexes=None,
        from_table=None,
    ):
        self.name = table_name
        self.new_name = False
        self.added_columns = added_columns or OrderedDict()
        self.changed_columns = changed_columns or OrderedDict()
        self.removed_columns = removed_columns or OrderedDict()
        self.added_indexes = added_indexes or OrderedDict()
        self.changed_indexes = changed_indexes or OrderedDict()
        self.removed_indexes = removed_indexes or OrderedDict()
        self.added_foreign_keys = []
        self.changed_foreign_keys = []
        self.removed_foreign_keys = []
        self.renamed_columns = OrderedDict()
        self.renamed_indexes = OrderedDict()
        self.from_table = from_table

    def get_name(self, platform):
        if isinstance(self.from_table, Table):
            name = self.from_table.get_quoted_name(platform)
        else:
            name = self.name

        return Identifier(name)

    def get_new_name(self):
        if self.new_name:
            return Identifier(self.new_name)

        return self.new_name



# -*- coding: utf-8 -*-

from .grammar import SchemaGrammar
from ..blueprint import Blueprint
from ...query.expression import QueryExpression
from ...support.fluent import Fluent


class MySQLSchemaGrammar(SchemaGrammar):

    _modifiers = [
        "unsigned",
        "charset",
        "collate",
        "nullable",
        "default",
        "increment",
        "comment",
        "after",
    ]

    _serials = [
        "big_integer",
        "integer",
        "medium_integer",
        "small_integer",
        "tiny_integer",
    ]

    marker = "%s"

    def compile_table_exists(self):
        """
        Compile the query to determine if a table exists

        :rtype: str
        """
        return (
            "SELECT * "
            "FROM information_schema.tables "
            "WHERE table_schema = %(marker)s "
            "AND table_name = %(marker)s" % {"marker": self.get_marker()}
        )

    def compile_column_exists(self):
        """
        Compile the query to determine the list of columns.
        """
        return (
            "SELECT column_name "
            "FROM information_schema.columns "
            "WHERE table_schema = %(marker)s AND table_name = %(marker)s"
            % {"marker": self.get_marker()}
        )

    def compile_create(self, blueprint, command, connection):
        """
        Compile a create table command.
        """
        columns = ", ".join(self._get_columns(blueprint))

        sql = "CREATE TABLE %s (%s)" % (self.wrap_table(blueprint), columns)

        sql = self._compile_create_encoding(sql, connection, blueprint)

        if blueprint.engine:
            sql += " ENGINE = %s" % blueprint.engine

        return sql

    def _compile_create_encoding(self, sql, connection, blueprint):
        """
        Append the character set specifications to a command.

        :type sql: str
        :type connection: orator.connections.Connection
        :type blueprint: Blueprint

        :rtype: str
        """
        charset = blueprint.charset or connection.get_config("charset")
        if charset:
            sql += " DEFAULT CHARACTER SET %s" % charset

        collation = blueprint.collation or connection.get_config("collation")
        if collation:
            sql += " COLLATE %s" % collation

        return sql

    def compile_add(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        columns = self.prefix_list("ADD", self._get_columns(blueprint))

        return "ALTER TABLE %s %s" % (table, ", ".join(columns))

    def compile_primary(self, blueprint, command, _):
        command.name = None

        return self._compile_key(blueprint, command, "PRIMARY KEY")

    def compile_unique(self, blueprint, command, _):
        return self._compile_key(blueprint, command, "UNIQUE")

    def compile_index(self, blueprint, command, _):
        return self._compile_key(blueprint, command, "INDEX")

    def _compile_key(self, blueprint, command, type):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s ADD %s %s(%s)" % (table, type, command.index, columns)

    def compile_drop(self, blueprint, command, _):
        return "DROP TABLE %s" % self.wrap_table(blueprint)

    def compile_drop_if_exists(self, blueprint, command, _):
        return "DROP TABLE IF EXISTS %s" % self.wrap_table(blueprint)

    def compile_drop_column(self, blueprint, command, connection):
        columns = self.prefix_list("DROP", self.wrap_list(command.columns))

        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s %s" % (table, ", ".join(columns))

    def compile_drop_primary(self, blueprint, command, _):
        return "ALTER TABLE %s DROP PRIMARY KEY" % self.wrap_table(blueprint)

    def compile_drop_unique(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s DROP INDEX %s" % (table, command.index)

    def compile_drop_index(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s DROP INDEX %s" % (table, command.index)

    def compile_drop_foreign(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s DROP FOREIGN KEY %s" % (table, command.index)

    def compile_rename(self, blueprint, command, _):
        from_ = self.wrap_table(blueprint)

        return "RENAME TABLE %s TO %s" % (from_, self.wrap_table(command.to))

    def _type_char(self, column):
        return "CHAR(%s)" % column.length

    def _type_string(self, column):
        return "VARCHAR(%s)" % column.length

    def _type_text(self, column):
        return "TEXT"

    def _type_medium_text(self, column):
        return "MEDIUMTEXT"

    def _type_long_text(self, column):
        return "LONGTEXT"

    def _type_integer(self, column):
        return "INT"

    def _type_big_integer(self, column):
        return "BIGINT"

    def _type_medium_integer(self, column):
        return "MEDIUMINT"

    def _type_tiny_integer(self, column):
        return "TINYINT"

    def _type_small_integer(self, column):
        return "SMALLINT"

    def _type_float(self, column):
        return self._type_double(column)

    def _type_double(self, column):
        if column.total and column.places:
            return "DOUBLE(%s, %s)" % (column.total, column.places)

        return "DOUBLE"

    def _type_decimal(self, column):
        return "DECIMAL(%s, %s)" % (column.total, column.places)

    def _type_boolean(self, column):
        return "TINYINT(1)"

    def _type_enum(self, column):
        return "ENUM('%s')" % "', '".join(column.allowed)

    def _type_json(self, column):
        if self.platform().has_native_json_type():
            return "JSON"

        return "TEXT"

    def _type_date(self, column):
        return "DATE"

    def _type_datetime(self, column):
        return "DATETIME"

    def _type_time(self, column):
        return "TIME"

    def _type_timestamp(self, column):
        platform_version = self.platform_version(3)
        column_type = "TIMESTAMP"

        if platform_version >= (5, 6, 0):
            if platform_version >= (5, 6, 4):
                # Versions 5.6.4+ support fractional seconds
                column_type = "TIMESTAMP(6)"
                current = "CURRENT_TIMESTAMP(6)"
            else:
                current = "CURRENT_TIMESTAMP"
        else:
            current = "0"

        if column.use_current:
            return "{} DEFAULT {}".format(column_type, current)

        return column_type

    def _type_binary(self, column):
        return "BLOB"

    def _modify_unsigned(self, blueprint, column):
        if column.get("unsigned", False):
            return " UNSIGNED"

        return ""

    def _modify_charset(self, blueprint, column):
        if column.get("charset"):
            return " CHARACTER SET " + column.charset

        return ""

    def _modify_collate(self, blueprint, column):
        if column.get("collation"):
            return " COLLATE " + column.collation

        return ""

    def _modify_nullable(self, blueprint, column):
        if column.get("nullable"):
            return " NULL"

        return " NOT NULL"

    def _modify_default(self, blueprint, column):
        if column.get("default") is not None:
            return " DEFAULT %s" % self._get_default_value(column.default)

        return ""

    def _modify_increment(self, blueprint, column):
        if column.type in self._serials and column.auto_increment:
            return " AUTO_INCREMENT PRIMARY KEY"

        return ""

    def _modify_after(self, blueprint, column):
        if column.get("after") is not None:
            return " AFTER " + self.wrap(column.after)

        return ""

    def _modify_comment(self, blueprint, column):
        if column.get("comment") is not None:
            return ' COMMENT "%s"' % column.comment

        return ""

    def _get_column_change_options(self, fluent):
        """
        Get the column change options.
        """
        options = super(MySQLSchemaGrammar, self)._get_column_change_options(fluent)

        if fluent.type == "enum":
            options["extra"] = {
                "definition": "('{}')".format("','".join(fluent.allowed))
            }

        return options

    def _wrap_value(self, value):
        if value == "*":
            return value

        return "`%s`" % value.replace("`", "``")



# -*- coding: utf-8 -*-

import importlib
import inflection
import os
from cleo import InputOption
from orator import DatabaseManager
from .base_command import BaseCommand
from ...utils import load_module


class SeedCommand(BaseCommand):
    """
    Seed the database with records.

    db:seed
        {--d|database= : The database connection to use.}
        {--p|path= : The path to seeders files.
                     Defaults to <comment>./seeds</comment>.}
        {--seeder=database_seeder : The name of the root seeder.}
        {--f|force : Force the operation to run.}
    """

    def handle(self):
        """
        Executes the command.
        """
        if not self.confirm_to_proceed(
            "<question>Are you sure you want to seed the database?:</question> "
        ):
            return

        self.resolver.set_default_connection(self.option("database"))

        self._get_seeder().run()

        self.info("Database seeded!")

    def _get_seeder(self):
        name = self._parse_name(self.option("seeder"))
        seeder_file = self._get_path(name)

        # Loading parent module
        load_module("seeds", self._get_path("__init__"))

        # Loading module
        mod = load_module("seeds.%s" % name, seeder_file)

        klass = getattr(mod, inflection.camelize(name))

        instance = klass()
        instance.set_command(self)
        instance.set_connection_resolver(self.resolver)

        return instance

    def _parse_name(self, name):
        if name.endswith(".py"):
            name = name.replace(".py", "", -1)

        return name

    def _get_path(self, name):
        """
        Get the destination class path.

        :param name: The name
        :type name: str

        :rtype: str
        """
        path = self.option("path")
        if path is None:
            path = self._get_seeders_path()

        return os.path.join(path, "%s.py" % name)



# -*- coding: utf-8 -*-

from .builder import SchemaBuilder


class MySQLSchemaBuilder(SchemaBuilder):
    def has_table(self, table):
        """
        Determine if the given table exists.

        :param table: The table
        :type table: str

        :rtype: bool
        """
        sql = self._grammar.compile_table_exists()
        database = self._connection.get_database_name()
        table = self._connection.get_table_prefix() + table

        return len(self._connection.select(sql, [database, table])) > 0

    def get_column_listing(self, table):
        """
        Get the column listing for a given table.

        :param table: The table
        :type table: str

        :rtype: list
        """
        sql = self._grammar.compile_column_exists()
        database = self._connection.get_database_name()
        table = self._connection.get_table_prefix() + table

        results = []
        for result in self._connection.select(sql, [database, table]):
            new_result = {}
            for key, value in result.items():
                new_result[key.lower()] = value

            results.append(new_result)

        return self._connection.get_post_processor().process_column_listing(results)



# -*- coding: utf-8 -*-

from ..orm import Factory


class Seeder(object):

    factory = None

    def __init__(self, resolver=None):
        self._command = None
        self._resolver = resolver

        if self.factory is None:
            self.factory = Factory(resolver=resolver)
        else:
            self.factory.set_connection_resolver(self._resolver)

    def run(self):
        """
        Run the database seeds.
        """
        pass

    def call(self, klass):
        """
        Seed the given connection from the given class.

        :param klass: The Seeder class
        :type klass: class
        """
        self._resolve(klass).run()

        if self._command:
            self._command.line("<info>Seeded:</info> <fg=cyan>%s</>" % klass.__name__)

    def _resolve(self, klass):
        """
        Resolve an instance of the given seeder klass.

        :param klass: The Seeder class
        :type klass: class
        """
        resolver = None

        if self._resolver:
            resolver = self._resolver
        elif self._command:
            resolver = self._command.resolver

        instance = klass()
        instance.set_connection_resolver(resolver)

        if self._command:
            instance.set_command(self._command)

        return instance

    def set_command(self, command):
        """
        Set the console command instance.

        :param command: The command
        :type command: cleo.Command
        """
        self._command = command

        return self

    def set_connection_resolver(self, resolver):
        self._resolver = resolver
        self.factory.set_connection_resolver(resolver)

    @property
    def db(self):
        return self._resolver



# -*- coding: utf-8 -*-

import os
import inflection
from faker import Faker
from functools import wraps
from .factory_builder import FactoryBuilder


class Factory(object):
    def __init__(self, faker=None, resolver=None):
        """
        :param faker: A faker generator instance
        :type faker: faker.Generator
        """
        if faker is None:
            self._faker = Faker()
        else:
            self._faker = faker

        self._definitions = {}
        self._resolver = resolver

    @classmethod
    def construct(cls, faker, path_to_factories=None):
        """
        Create a new factory container.

        :param faker: A faker generator instance
        :type faker: faker.Generator

        :param path_to_factories: The path to factories
        :type path_to_factories: str

        :rtype: Factory
        """
        factory = faker.__class__()

        if path_to_factories is not None and os.path.isdir(path_to_factories):
            for filename in os.listdir(path_to_factories):
                if os.path.isfile(filename):
                    cls._resolve(path_to_factories, filename)

        return factory

    def define_as(self, klass, name):
        """
        Define a class with the given short name.

        :param klass: The class
        :type klass: class

        :param name: The short name
        :type name: str
        """
        return self.define(klass, name)

    def define(self, klass, name="default"):
        """
        Define a class with a given set of attributes.

        :param klass: The class
        :type klass: class

        :param name: The short name
        :type name: str
        """

        def decorate(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)

            self.register(klass, func, name=name)

            return wrapped

        return decorate

    def register(self, klass, callback, name="default"):
        """
        Register a class with a function.

        :param klass: The class
        :type klass: class

        :param callback: The callable
        :type callback: callable

        :param name: The short name
        :type name: str
        """
        if klass not in self._definitions:
            self._definitions[klass] = {}

        self._definitions[klass][name] = callback

    def register_as(self, klass, name, callback):
        """
        Register a class with a function.

        :param klass: The class
        :type klass: class

        :param callback: The callable
        :type callback: callable

        :param name: The short name
        :type name: str
        """
        return self.register(klass, callback, name)

    def create(self, klass, **attributes):
        """
        Create an instance of the given model and persist it to the database.

        :param klass: The class
        :type klass: class

        :param attributes: The instance attributes
        :type attributes: dict

        :return: mixed
        """
        return self.of(klass).create(**attributes)

    def create_as(self, klass, name, **attributes):
        """
        Create an instance of the given model and type and persist it to the database.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :param attributes: The instance attributes
        :type attributes: dict

        :return: mixed
        """
        return self.of(klass, name).create(**attributes)

    def make(self, klass, **attributes):
        """
        Create an instance of the given model.

        :param klass: The class
        :type klass: class

        :param attributes: The instance attributes
        :type attributes: dict

        :return: mixed
        """
        return self.of(klass).make(**attributes)

    def make_as(self, klass, name, **attributes):
        """
        Create an instance of the given model and type.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :param attributes: The instance attributes
        :type attributes: dict

        :return: mixed
        """
        return self.of(klass, name).make(**attributes)

    def raw_of(self, klass, name, **attributes):
        """
        Get the raw attribute dict for a given named model.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :param attributes: The instance attributes
        :type attributes: dict

        :return: dict
        """
        return self.raw(klass, _name=name, **attributes)

    def raw(self, klass, _name="default", **attributes):
        """
        Get the raw attribute dict for a given named model.

        :param klass: The class
        :type klass: class

        :param _name: The type
        :type _name: str

        :param attributes: The instance attributes
        :type attributes: dict

        :return: dict
        """
        raw = self._definitions[klass][_name](self._faker)

        raw.update(attributes)

        return raw

    def of(self, klass, name="default"):
        """
        Create a builder for the given model.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :return: orator.orm.factory_builder.FactoryBuilder
        """
        return FactoryBuilder(
            klass, name, self._definitions, self._faker, self._resolver
        )

    def build(self, klass, name="default", amount=None):
        """
        Makes a factory builder with a specified amount.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :param amount: The number of models to create
        :type amount: int

        :return: mixed
        """
        if amount is None:
            if isinstance(name, int):
                amount = name
                name = "default"
            else:
                amount = 1

        return self.of(klass, name).times(amount)

    @classmethod
    def _resolve(cls, path, factory_file):
        """
        Resolve a migration instance from a file.

        :param path: The path to factories directory
        :type path: str

        :param factory_file: The migration file
        :type factory_file: str

        :rtype: Factory
        """
        variables = {}

        name = factory_file
        factory_file = os.path.join(path, factory_file)

        with open(factory_file) as fh:
            exec(fh.read(), {}, variables)

        klass = variables[inflection.camelize(name)]

        instance = klass()

        return instance

    def set_connection_resolver(self, resolver):
        self._resolver = resolver

    def __getitem__(self, item):
        return self.make(item)

    def __setitem__(self, key, value):
        return self.define(key, value)

    def __contains__(self, item):
        return item in self._definitions

    def __call__(self, klass, name="default", amount=None):
        """
        Makes a factory builder with a specified amount.

        :param klass: The class
        :type klass: class

        :param name: The type
        :type name: str

        :param amount: The number of models to create
        :type amount: int

        :return: mixed
        """
        return self.build(klass, name, amount)



# -*- coding: utf-8 -*-

from lazy_object_proxy import Proxy
from functools import wraps


def wrapped(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        return Wrapper(func(*args, **kwargs))

    return _wrapper


class Wrapper(Proxy):
    """
    Wrapper around a relation which provide
    dynamic property functionality.
    """

    _relation = None

    def __init__(self, relation):
        """
        :param relation: The underlying relation.
        :type relation: Relation
        :return:
        """
        super(Wrapper, self).__init__(self._get_results)

        self._relation = relation

    def _get_results(self):
        return self._relation.get_results()

    def __call__(self, *args, **kwargs):
        return self._relation.new_instance(self._relation.get_parent())

    def __repr__(self):
        return repr(self.__wrapped__)


class BelongsToManyWrapper(Wrapper):
    def with_timestamps(self):
        self._relation.with_timestamps()

        return self

    def with_pivot(self, *columns):
        self._relation.with_pivot(*columns)

        return self



# -*- coding: utf-8 -*-

from .expression import QueryExpression


class JoinClause(object):
    def __init__(self, table, type="inner"):
        self.type = type
        self.table = table

        self.clauses = []
        self.bindings = []

    def on(self, first, operator, second, boolean="and", where=False):
        self.clauses.append(
            {
                "first": first,
                "operator": operator,
                "second": second,
                "boolean": boolean,
                "where": where,
            }
        )

        if where:
            self.bindings.append(second)

        return self

    def or_on(self, first, operator, second):
        return self.on(first, operator, second, "or")

    def where(self, first, operator, second, boolean="and"):
        return self.on(first, operator, second, boolean, True)

    def or_where(self, first, operator, second):
        return self.where(first, operator, second, "or")

    def where_null(self, column, boolean="and"):
        return self.on(column, "IS", QueryExpression("NULL"), boolean, False)

    def or_where_null(self, column):
        return self.where_null(column, "or")

    def where_not_null(self, column, boolean="and"):
        return self.on(column, "IS", QueryExpression("NOT NULL"), boolean, False)

    def or_where_not_null(self, column):
        return self.where_not_null(column, "or")



# -*- coding: utf-8 -*-

from cleo import Application
from .. import __version__

application = Application("Orator", __version__, complete=True)

# Migrations
from .migrations import (
    InstallCommand,
    MigrateCommand,
    MigrateMakeCommand,
    RollbackCommand,
    StatusCommand,
    ResetCommand,
    RefreshCommand,
)

application.add(InstallCommand())
application.add(MigrateCommand())
application.add(MigrateMakeCommand())
application.add(RollbackCommand())
application.add(StatusCommand())
application.add(ResetCommand())
application.add(RefreshCommand())

# Seeds
from .seeds import SeedersMakeCommand, SeedCommand

application.add(SeedersMakeCommand())
application.add(SeedCommand())

# Models
from .models import ModelMakeCommand

application.add(ModelMakeCommand())



# -*- coding: utf-8 -*-

import hashlib
import time
import inflection
from ...exceptions.orm import ModelNotFound
from ...query.expression import QueryExpression
from ..collection import Collection
import orator.orm.model
from .relation import Relation
from .result import Result


class BelongsToMany(Relation):

    _table = None
    _other_key = None
    _foreign_key = None
    _relation_name = None

    _pivot_columns = []
    _pivot_wheres = []

    def __init__(
        self, query, parent, table, foreign_key, other_key, relation_name=None
    ):
        """
        :param query: A Builder instance
        :type query: Builder

        :param parent: The parent model
        :type parent: Model

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :param relation_name: The relation name
        :type relation_name: str
        """
        self._table = table
        self._other_key = other_key
        self._foreign_key = foreign_key
        self._relation_name = relation_name

        self._pivot_columns = []
        self._pivot_wheres = []

        super(BelongsToMany, self).__init__(query, parent)

    def get_results(self):
        """
        Get the results of the relationship.
        """
        return self.get()

    def where_pivot(self, column, operator=None, value=None, boolean="and"):
        """
        Set a where clause for a pivot table column.

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str|Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :param boolean: The boolean of the where clause
        :type boolean: str

        :return: self
        :rtype: self
        """
        self._pivot_wheres.append([column, operator, value, boolean])

        return self._query.where(
            "%s.%s" % (self._table, column), operator, value, boolean
        )

    def or_where_pivot(self, column, operator=None, value=None):
        """
        Set an or where clause for a pivot table column.

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str|Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :return: self
        :rtype: BelongsToMany
        """
        return self.where_pivot(column, operator, value, "or")

    def first(self, columns=None):
        """
        Execute the query and get the first result.

        :type columns: list
        """
        self._query.take(1)

        results = self.get(columns)

        if len(results) > 0:
            return results.first()

        return

    def first_or_fail(self, columns=None):
        """
        Execute the query and get the first result or raise an exception.

        :type columns: list

        :raises: ModelNotFound
        """
        model = self.first(columns)
        if model is not None:
            return model

        raise ModelNotFound(self._parent.__class__)

    def get(self, columns=None):
        """
        Execute the query as a "select" statement.

        :type columns: list

        :rtype: orator.Collection
        """
        if columns is None:
            columns = ["*"]

        if self._query.get_query().columns:
            columns = []

        select = self._get_select_columns(columns)

        models = self._query.add_select(*select).get_models()

        self._hydrate_pivot_relation(models)

        if len(models) > 0:
            models = self._query.eager_load_relations(models)

        return self._related.new_collection(models)

    def _hydrate_pivot_relation(self, models):
        """
        Hydrate the pivot table relationship on the models.

        :type models: list
        """
        for model in models:
            pivot = self.new_existing_pivot(self._clean_pivot_attributes(model))

            model.set_relation("pivot", pivot)

    def _clean_pivot_attributes(self, model):
        """
        Get the pivot attributes from a model.

        :type model: orator.Model
        """
        values = {}
        delete_keys = []

        for key, value in model.get_attributes().items():
            if key.find("pivot_") == 0:
                values[key[6:]] = value

                delete_keys.append(key)

        for key in delete_keys:
            delattr(model, key)

        return values

    def add_constraints(self):
        """
        Set the base constraints on the relation query.

        :rtype: None
        """
        self._set_join()

        if BelongsToMany._constraints:
            self._set_where()

    def get_relation_count_query(self, query, parent):
        """
        Add the constraints for a relationship count query.

        :type query: orator.orm.Builder
        :type parent: orator.orm.Builder

        :rtype: orator.orm.Builder
        """
        if parent.get_query().from__ == query.get_query().from__:
            return self.get_relation_count_query_for_self_join(query, parent)

        self._set_join(query)

        return super(BelongsToMany, self).get_relation_count_query(query, parent)

    def get_relation_count_query_for_self_join(self, query, parent):
        """
        Add the constraints for a relationship count query on the same table.

        :type query: orator.orm.Builder
        :type parent: orator.orm.Builder

        :rtype: orator.orm.Builder
        """
        query.select(QueryExpression("COUNT(*)"))

        table_prefix = self._query.get_query().get_connection().get_table_prefix()

        hash_ = self.get_relation_count_hash()
        query.from_("%s AS %s%s" % (self._table, table_prefix, hash_))

        key = self.wrap(self.get_qualified_parent_key_name())

        return query.where(
            "%s.%s" % (hash_, self._foreign_key), "=", QueryExpression(key)
        )

    def get_relation_count_hash(self):
        """
        Get a relationship join table hash.

        :rtype: str
        """
        return "self_%s" % (hashlib.md5(str(time.time()).encode()).hexdigest())

    def _get_select_columns(self, columns=None):
        """
        Set the select clause for the relation query.

        :param columns: The columns
        :type columns: list

        :rtype: list
        """
        if columns == ["*"] or columns is None:
            columns = ["%s.*" % self._related.get_table()]

        return columns + self._get_aliased_pivot_columns()

    def _get_aliased_pivot_columns(self):
        """
        Get the pivot columns for the relation.

        :rtype: list
        """
        defaults = [self._foreign_key, self._other_key]

        columns = []

        for column in defaults + self._pivot_columns:
            value = "%s.%s AS pivot_%s" % (self._table, column, column)
            if value not in columns:
                columns.append("%s.%s AS pivot_%s" % (self._table, column, column))

        return columns

    def _has_pivot_column(self, column):
        """
        Determine whether the given column is defined as a pivot column.

        :param column: The column to check
        :type column: str

        :rtype: bool
        """
        return column in self._pivot_columns

    def _set_join(self, query=None):
        """
        Set the join clause for the relation query.

        :param query: The query builder
        :type query: orator.orm.Builder

        :return: self
        :rtype: BelongsToMany
        """
        if not query:
            query = self._query

        base_table = self._related.get_table()

        key = "%s.%s" % (base_table, self._related.get_key_name())

        query.join(self._table, key, "=", self.get_other_key())

        return self

    def _set_where(self):
        """
        Set the where clause for the relation query.

        :return: self
        :rtype: BelongsToMany
        """
        foreign = self.get_foreign_key()

        self._query.where(foreign, "=", self._parent.get_key())

        return self

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        self._query.where_in(self.get_foreign_key(), self.get_keys(models))

    def init_relation(self, models, relation):
        """
        Initialize the relation on a set of models.

        :type models: list
        :type relation:  str
        """
        for model in models:
            model.set_relation(
                relation, Result(self._related.new_collection(), self, model)
            )

        return models

    def match(self, models, results, relation):
        """
        Match the eagerly loaded results to their parents.

        :type models: list
        :type results: Collection
        :type relation:  str
        """
        dictionary = self._build_dictionary(results)

        for model in models:
            key = model.get_key()

            if key in dictionary:
                collection = Result(
                    self._related.new_collection(dictionary[key]), self, model
                )
            else:
                collection = Result(self._related.new_collection(), self, model)

            model.set_relation(relation, collection)

        return models

    def _build_dictionary(self, results):
        """
        Build model dictionary keyed by the relation's foreign key.

        :param results: The results
        :type results: Collection

        :rtype: dict
        """
        foreign = self._foreign_key

        dictionary = {}

        for result in results:
            key = getattr(result.pivot, foreign)
            if key not in dictionary:
                dictionary[key] = []

            dictionary[key].append(result)

        return dictionary

    def touch(self):
        """
        Touch all of the related models of the relationship.
        """
        key = self.get_related().get_key_name()

        columns = self.get_related_fresh_update()

        ids = self.get_related_ids()

        if len(ids) > 0:
            self.get_related().new_query().where_in(key, ids).update(columns)

    def get_related_ids(self):
        """
        Get all of the IDs for the related models.

        :rtype: list
        """
        related = self.get_related()

        full_key = related.get_qualified_key_name()

        return self.get_query().select(full_key).lists(related.get_key_name())

    def save(self, model, joining=None, touch=True):
        """
        Save a new model and attach it to the parent model.

        :type model: orator.Model
        :type joining: dict
        :type touch: bool

        :rtype: orator.Model
        """
        if joining is None:
            joining = {}

        model.save({"touch": False})

        self.attach(model.get_key(), joining, touch)

        return model

    def save_many(self, models, joinings=None):
        """
        Save a list of new models and attach them to the parent model

        :type models: list
        :type joinings: dict

        :rtype: list
        """
        if joinings is None:
            joinings = {}

        for key, model in enumerate(models):
            self.save(model, joinings.get(key), False)

        self.touch_if_touching()

        return models

    def find_or_new(self, id, columns=None):
        """
        Find a model by its primary key or return new instance of the related model.

        :param id: The primary key
        :type id: mixed

        :param columns:  The columns to retrieve
        :type columns: list

        :rtype: Collection or Model
        """
        instance = self._query.find(id, columns)
        if instance is None:
            instance = self.get_related().new_instance()

        return instance

    def first_or_new(self, _attributes=None, **attributes):
        """
        Get the first related model record matching the attributes or instantiate it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._query.where(attributes).first()
        if instance is None:
            instance = self._related.new_instance()

        return instance

    def first_or_create(
        self, _attributes=None, _joining=None, _touch=True, **attributes
    ):
        """
        Get the first related model record matching the attributes or create it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._query.where(attributes).first()
        if instance is None:
            instance = self.create(attributes, _joining or {}, _touch)

        return instance

    def update_or_create(self, attributes, values=None, joining=None, touch=True):
        """
        Create or update a related record matching the attributes, and fill it with values.

        :param attributes: The attributes
        :type attributes: dict

        :param values: The values
        :type values: dict

        :rtype: Model
        """
        if values is None:
            values = {}

        instance = self._query.where(attributes).first()

        if instance is None:
            return self.create(values, joining, touch)

        instance.fill(**values)

        instance.save({"touch": False})

        return instance

    def create(self, _attributes=None, _joining=None, _touch=True, **attributes):
        """
        Create a new instance of the related model.

        :param attributes: The attributes
        :type attributes: dict

        :rtype: orator.orm.Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._related.new_instance(attributes)

        instance.save({"touch": False})

        self.attach(instance.get_key(), _joining, _touch)

        return instance

    def create_many(self, records, joinings=None):
        """
        Create a list of new instances of the related model.
        """
        if joinings is None:
            joinings = []

        instances = []

        for key, record in enumerate(records):
            instances.append(self.create(record), joinings[key], False)

        self.touch_if_touching()

        return instances

    def sync(self, ids, detaching=True):
        """
        Sync the intermediate tables with a list of IDs or collection of models
        """
        changes = {"attached": [], "detached": [], "updated": []}

        if isinstance(ids, Collection):
            ids = ids.model_keys()

        current = self._new_pivot_query().lists(self._other_key).all()

        records = self._format_sync_list(ids)

        detach = [x for x in current if x not in records.keys()]

        if detaching and len(detach) > 0:
            self.detach(detach)

            changes["detached"] = detach

        changes.update(self._attach_new(records, current, False))

        if len(changes["attached"]) or len(changes["updated"]):
            self.touch_if_touching()

        return changes

    def _format_sync_list(self, records):
        """
        Format the sync list so that it is keyed by ID.
        """
        results = {}

        for attributes in records:
            if not isinstance(attributes, dict):
                id, attributes = attributes, {}
            else:
                id = list(attributes.keys())[0]
                attributes = attributes[id]

            results[id] = attributes

        return results

    def _attach_new(self, records, current, touch=True):
        """
        Attach all of the IDs that aren't in the current dict.
        """
        changes = {"attached": [], "updated": []}

        for id, attributes in records.items():
            if id not in current:
                self.attach(id, attributes, touch)

                changes["attached"].append(id)
            elif len(attributes) > 0 and self.update_existing_pivot(
                id, attributes, touch
            ):
                changes["updated"].append(id)

        return changes

    def update_existing_pivot(self, id, attributes, touch=True):
        """
        Update an existing pivot record on the table.
        """
        if self.updated_at() in self._pivot_columns:
            attributes = self.set_timestamps_on_attach(attributes, True)

        updated = self.new_pivot_statement_for_id(id).update(attributes)

        if touch:
            self.touch_if_touching()

        return updated

    def attach(self, id, attributes=None, touch=True):
        """
        Attach a model to the parent.
        """
        if isinstance(id, orator.orm.Model):
            id = id.get_key()

        query = self.new_pivot_statement()

        if not isinstance(id, list):
            id = [id]

        query.insert(self._create_attach_records(id, attributes))

        if touch:
            self.touch_if_touching()

    def _create_attach_records(self, ids, attributes):
        """
        Create a list of records to insert into the pivot table.
        """
        records = []

        timed = self._has_pivot_column(self.created_at()) or self._has_pivot_column(
            self.updated_at()
        )

        for key, value in enumerate(ids):
            records.append(self._attacher(key, value, attributes, timed))

        return records

    def _attacher(self, key, value, attributes, timed):
        """
        Create a full attachment record payload.
        """
        id, extra = self._get_attach_id(key, value, attributes)

        record = self._create_attach_record(id, timed)

        if extra:
            record.update(extra)

        return record

    def _get_attach_id(self, key, value, attributes):
        """
        Get the attach record ID and extra attributes.
        """
        if isinstance(value, dict):
            key = list(value.keys())[0]
            attributes.update(value[key])

            return [key, attributes]

        return value, attributes

    def _create_attach_record(self, id, timed):
        """
        Create a new pivot attachement record.
        """
        record = {}

        record[self._foreign_key] = self._parent.get_key()

        record[self._other_key] = id

        if timed:
            record = self._set_timestamps_on_attach(record)

        return record

    def _set_timestamps_on_attach(self, record, exists=False):
        """
        Set the creation an update timestamps on an attach record.
        """
        fresh = self._parent.fresh_timestamp()

        if not exists and self._has_pivot_column(self.created_at()):
            record[self.created_at()] = fresh

        if self._has_pivot_column(self.updated_at()):
            record[self.updated_at()] = fresh

        return record

    def detach(self, ids=None, touch=True):
        """
        Detach models from the relationship.
        """
        if isinstance(ids, orator.orm.model.Model):
            ids = ids.get_key()

        if ids is None:
            ids = []

        query = self._new_pivot_query()

        if not isinstance(ids, list):
            ids = [ids]

        if len(ids) > 0:
            query.where_in(self._other_key, ids)

        if touch:
            self.touch_if_touching()

        results = query.delete()

        return results

    def touch_if_touching(self):
        """
        Touch if the parent model is being touched.
        """
        if self._touching_parent():
            self.get_parent().touch()

        if self.get_parent().touches(self._relation_name):
            self.touch()

    def _touching_parent(self):
        """
        Determine if we should touch the parent on sync.
        """
        return self.get_related().touches(self._guess_inverse_relation())

    def _guess_inverse_relation(self):
        return inflection.camelize(
            inflection.pluralize(self.get_parent().__class__.__name__)
        )

    def _new_pivot_query(self):
        """
        Create a new query builder for the pivot table.

        :rtype: orator.orm.Builder
        """
        query = self.new_pivot_statement()

        for where_args in self._pivot_wheres:
            query.where(*where_args)

        return query.where(self._foreign_key, self._parent.get_key())

    def new_pivot_statement(self):
        """
        Get a new plain query builder for the pivot table.
        """
        return self._query.get_query().new_query().from_(self._table)

    def new_pivot_statement_for_id(self, id):
        """
        Get a new pivot statement for a given "other" id.
        """
        return self._new_pivot_query().where(self._other_key, id)

    def new_pivot(self, attributes=None, exists=False):
        """
        Create a new pivot model instance.
        """
        pivot = self._related.new_pivot(self._parent, attributes, self._table, exists)

        return pivot.set_pivot_keys(self._foreign_key, self._other_key)

    def new_existing_pivot(self, attributes):
        """
        Create a new existing pivot model instance.
        """
        return self.new_pivot(attributes, True)

    def with_pivot(self, *columns):
        """
        Set the columns on the pivot table to retrieve.
        """
        columns = list(columns)

        self._pivot_columns += columns

        return self

    def with_timestamps(self, created_at=None, updated_at=None):
        """
        Specify that the pivot table has creation and update columns.
        """
        if not created_at:
            created_at = self.created_at()

        if not updated_at:
            updated_at = self.updated_at()

        return self.with_pivot(created_at, updated_at)

    def get_related_fresh_update(self):
        """
        Get the related model's update at column at
        """
        return {self._related.get_updated_at_column(): self._related.fresh_timestamp()}

    def get_has_compare_key(self):
        """
        Get the key for comparing against the parent key in "has" query.
        """
        return self.get_foreign_key()

    def get_foreign_key(self):
        return "%s.%s" % (self._table, self._foreign_key)

    def get_other_key(self):
        return "%s.%s" % (self._table, self._other_key)

    def get_table(self):
        return self._table

    def get_relation_name(self):
        return self._relation_name

    def _new_instance(self, model):
        relation = BelongsToMany(
            self.new_query(),
            model,
            self._table,
            self._foreign_key,
            self._other_key,
            self._relation_name,
        )

        relation.with_pivot(*self._pivot_columns)

        return relation



# -*- coding: utf-8 -*-

import random
from ..exceptions import ArgumentError
from ..exceptions.connectors import UnsupportedDriver
from .mysql_connector import MySQLConnector
from .postgres_connector import PostgresConnector
from .sqlite_connector import SQLiteConnector
from ..connections import MySQLConnection, PostgresConnection, SQLiteConnection


class ConnectionFactory(object):

    CONNECTORS = {
        "sqlite": SQLiteConnector,
        "mysql": MySQLConnector,
        "postgres": PostgresConnector,
        "pgsql": PostgresConnector,
    }

    CONNECTIONS = {
        "sqlite": SQLiteConnection,
        "mysql": MySQLConnection,
        "postgres": PostgresConnection,
        "pgsql": PostgresConnection,
    }

    def make(self, config, name=None):
        if "read" in config:
            return self._create_read_write_connection(config)

        return self._create_single_connection(config)

    def _create_single_connection(self, config):
        conn = self.create_connector(config).connect(config)

        return self._create_connection(
            config["driver"], conn, config["database"], config.get("prefix", ""), config
        )

    def _create_read_write_connection(self, config):
        connection = self._create_single_connection(self._get_write_config(config))

        connection.set_read_connection(self._create_read_connection(config))

        return connection

    def _create_read_connection(self, config):
        read_config = self._get_read_config(config)

        return self.create_connector(read_config).connect(read_config)

    def _get_read_config(self, config):
        read_config = self._get_read_write_config(config, "read")

        return self._merge_read_write_config(config, read_config)

    def _get_write_config(self, config):
        write_config = self._get_read_write_config(config, "write")

        return self._merge_read_write_config(config, write_config)

    def _get_read_write_config(self, config, type):
        if config.get(type, []):
            return random.choice(config[type])

        return config[type]

    def _merge_read_write_config(self, config, merge):
        config = config.copy()
        config.update(merge)

        del config["read"]
        del config["write"]

        return config

    def create_connector(self, config):
        if "driver" not in config:
            raise ArgumentError("A driver must be specified")

        driver = config["driver"]

        if driver not in self.CONNECTORS:
            raise UnsupportedDriver(driver)

        return self.CONNECTORS[driver](driver)

    @classmethod
    def register_connector(cls, name, connector):
        cls.CONNECTORS[connector] = connector

    @classmethod
    def register_connection(cls, name, connection):
        cls.CONNECTIONS[name] = connection

    def _create_connection(self, driver, connection, database, prefix="", config=None):
        if config is None:
            config = {}

        if driver not in self.CONNECTIONS:
            raise UnsupportedDriver(driver)

        return self.CONNECTIONS[driver](connection, database, prefix, config)
